【问题标题】:Replicate a vector multiple times using CUDA Thrust使用 CUDA Thrust 多次复制向量
【发布时间】:2015-09-08 16:27:30
【问题描述】:

我正在尝试使用 CUDA Thrust 解决问题。

我有一个带有 3 元素的主机数组。是否可以使用 Thrust 创建一个包含384 元素的设备数组,其中我的主机数组中的3 元素重复128 次(128 x 3 = 384)?

一般来说,从3元素的数组开始,如何使用Thrust生成大小为X的设备数组,其中X = Y x 3,即Y是重复次数?

【问题讨论】:

  • 推力expand example 展示了一些可能有用的想法。你可以说得更详细点吗?如果您的主机阵列是{1, 2, 3},您想以{1, 1, 1, ..., 2, 2, 2, ..., 3, 3, 3, ...} 结尾,还是想要{1, 2, 3, 1, 2, 3, ...},或者您想要别的什么?
  • 我想要{1, 2, 3, 1, 2, 3, .... }

标签: cuda thrust


【解决方案1】:

一种可能的方法:

  1. 创建适当大小的设备向量
  2. 创建 3 个strided ranges,为最终输出(设备)向量中的每个元素位置 {1、2、3} 创建一个
  3. 使用thrust::fill 用适当的(主向量)元素{1、2、3}填充3 个跨步范围中的每一个

此代码是对跨步范围示例的简单修改,以进行演示。您可以将 REPS 定义更改为 128 以查看到 384 个输出元素的完整扩展:

#include <thrust/iterator/counting_iterator.h>
#include <thrust/iterator/transform_iterator.h>
#include <thrust/iterator/permutation_iterator.h>
#include <thrust/functional.h>

#include <thrust/fill.h>
#include <thrust/device_vector.h>
#include <thrust/host_vector.h>

// for printing
#include <thrust/copy.h>
#include <ostream>


#define STRIDE 3
#define REPS  15  // change to 128 if you like
#define DSIZE (STRIDE*REPS)

// this example illustrates how to make strided access to a range of values
// examples:
//   strided_range([0, 1, 2, 3, 4, 5, 6], 1) -> [0, 1, 2, 3, 4, 5, 6]
//   strided_range([0, 1, 2, 3, 4, 5, 6], 2) -> [0, 2, 4, 6]
//   strided_range([0, 1, 2, 3, 4, 5, 6], 3) -> [0, 3, 6]
//   ...

template <typename Iterator>
class strided_range
{
    public:

    typedef typename thrust::iterator_difference<Iterator>::type difference_type;

    struct stride_functor : public thrust::unary_function<difference_type,difference_type>
    {
        difference_type stride;

        stride_functor(difference_type stride)
            : stride(stride) {}

        __host__ __device__
        difference_type operator()(const difference_type& i) const
        {
            return stride * i;
        }
    };

    typedef typename thrust::counting_iterator<difference_type>                   CountingIterator;
    typedef typename thrust::transform_iterator<stride_functor, CountingIterator> TransformIterator;
    typedef typename thrust::permutation_iterator<Iterator,TransformIterator>     PermutationIterator;

    // type of the strided_range iterator
    typedef PermutationIterator iterator;

    // construct strided_range for the range [first,last)
    strided_range(Iterator first, Iterator last, difference_type stride)
        : first(first), last(last), stride(stride) {}

    iterator begin(void) const
    {
        return PermutationIterator(first, TransformIterator(CountingIterator(0), stride_functor(stride)));
    }

    iterator end(void) const
    {
        return begin() + ((last - first) + (stride - 1)) / stride;
    }

    protected:
    Iterator first;
    Iterator last;
    difference_type stride;
};

int main(void)
{
    thrust::host_vector<int> h_data(STRIDE);
    h_data[0] = 1;
    h_data[1] = 2;
    h_data[2] = 3;

    thrust::device_vector<int> data(DSIZE);

    typedef thrust::device_vector<int>::iterator Iterator;
    strided_range<Iterator> pos1(data.begin(), data.end(), STRIDE);
    strided_range<Iterator> pos2(data.begin()+1, data.end(), STRIDE);
    strided_range<Iterator> pos3(data.begin()+2, data.end(), STRIDE);

    thrust::fill(pos1.begin(), pos1.end(), h_data[0]);
    thrust::fill(pos2.begin(), pos2.end(), h_data[1]);
    thrust::fill(pos3.begin(), pos3.end(), h_data[2]);


    // print the generated data
    std::cout << "data: ";
    thrust::copy(data.begin(), data.end(), std::ostream_iterator<int>(std::cout, " "));  std::cout << std::endl;

    return 0;
}

【讨论】:

  • 谢谢,应该可以的。
【解决方案2】:

Robert Crovella 已经使用strided ranges 回答了这个问题。他还指出了使用expand operator的可能性。

下面,我提供了一个使用扩展运算符的工作示例。与使用跨步范围相反,它避免了 for 循环的需要。

#include <thrust/device_vector.h>
#include <thrust/gather.h>
#include <thrust/sequence.h>
#include <stdio.h>

using namespace thrust::placeholders;

/*************************************/
/* CONVERT LINEAR INDEX TO ROW INDEX */
/*************************************/
template <typename T>
struct linear_index_to_row_index : public thrust::unary_function<T,T> {

    T Ncols; // --- Number of columns

    __host__ __device__ linear_index_to_row_index(T Ncols) : Ncols(Ncols) {}

    __host__ __device__ T operator()(T i) { return i / Ncols; }
};

/*******************/
/* EXPAND OPERATOR */
/*******************/
template <typename InputIterator1, typename InputIterator2, typename OutputIterator>
OutputIterator expand(InputIterator1 first1,
                      InputIterator1 last1,
                      InputIterator2 first2,
                      OutputIterator output)
{
    typedef typename thrust::iterator_difference<InputIterator1>::type difference_type;

    difference_type input_size  = thrust::distance(first1, last1);
    difference_type output_size = thrust::reduce(first1, last1);

    // scan the counts to obtain output offsets for each input element
    thrust::device_vector<difference_type> output_offsets(input_size, 0);
    thrust::exclusive_scan(first1, last1, output_offsets.begin()); 

    // scatter the nonzero counts into their corresponding output positions
    thrust::device_vector<difference_type> output_indices(output_size, 0);
    thrust::scatter_if(thrust::counting_iterator<difference_type>(0), thrust::counting_iterator<difference_type>(input_size),
                       output_offsets.begin(), first1, output_indices.begin());

    // compute max-scan over the output indices, filling in the holes
    thrust::inclusive_scan(output_indices.begin(), output_indices.end(), output_indices.begin(), thrust::maximum<difference_type>());

    // gather input values according to index array (output = first2[output_indices])
    OutputIterator output_end = output; thrust::advance(output_end, output_size);
    thrust::gather(output_indices.begin(), output_indices.end(), first2, output);

    // return output + output_size
    thrust::advance(output, output_size);

    return output;
}

/**************************/
/* STRIDED RANGE OPERATOR */
/**************************/
template <typename Iterator>
class strided_range
{
    public:

    typedef typename thrust::iterator_difference<Iterator>::type difference_type;

    struct stride_functor : public thrust::unary_function<difference_type,difference_type>
    {
        difference_type stride;

        stride_functor(difference_type stride)
            : stride(stride) {}

        __host__ __device__
        difference_type operator()(const difference_type& i) const
        {
            return stride * i;
        }
    };

    typedef typename thrust::counting_iterator<difference_type>                   CountingIterator;
    typedef typename thrust::transform_iterator<stride_functor, CountingIterator> TransformIterator;
    typedef typename thrust::permutation_iterator<Iterator,TransformIterator>     PermutationIterator;

    // type of the strided_range iterator
    typedef PermutationIterator iterator;

    // construct strided_range for the range [first,last)
    strided_range(Iterator first, Iterator last, difference_type stride)
        : first(first), last(last), stride(stride) {}

    iterator begin(void) const
    {
        return PermutationIterator(first, TransformIterator(CountingIterator(0), stride_functor(stride)));
    }

    iterator end(void) const
    {
        return begin() + ((last - first) + (stride - 1)) / stride;
    }

    protected:
    Iterator first;
    Iterator last;
    difference_type stride;
};

/********/
/* MAIN */
/********/
int main(){

    /**************************/
    /* SETTING UP THE PROBLEM */
    /**************************/

    const int Nrows = 10;           // --- Number of objects
    const int Ncols =  3;           // --- Number of centroids  

    thrust::device_vector<int> d_sequence(Nrows * Ncols);
    thrust::device_vector<int> d_counts(Ncols, Nrows);
    thrust::sequence(d_sequence.begin(), d_sequence.begin() + Ncols);
    expand(d_counts.begin(), d_counts.end(), d_sequence.begin(), 
        thrust::make_permutation_iterator(
                                d_sequence.begin(),
                                thrust::make_transform_iterator(thrust::make_counting_iterator(0),(_1 % Nrows) * Ncols + _1 / Nrows)));

    printf("\n\nCentroid indices\n");
    for(int i = 0; i < Nrows; i++) {
        std::cout << " [ ";
        for(int j = 0; j < Ncols; j++)
            std::cout << d_sequence[i * Ncols + j] << " ";
        std::cout << "]\n";
    }

    return 0;
}

【讨论】:

    【解决方案3】:

    作为使用 CUDA Thrust 的一个明显更简单的替代方案,我在下面发布了一个在 CUDA 中实现经典 Matlab 的 meshgrid 函数的工作示例。

    在 Matlab 中

    x = [1 2 3];
    y = [4 5 6 7];
    [X, Y] = meshgrid(x, y);
    

    生产

    X =
    
         1     2     3
         1     2     3
         1     2     3
         1     2     3
    

    Y =
    
         4     4     4
         5     5     5
         6     6     6
         7     7     7
    

    X 正是x 数组的四倍复制,这是OP 的问题和Robert Crovella 答案的第一个猜测,而Y 是@ 的每个元素的三倍连续复制987654327@数组,这是罗伯特·克罗维拉答案的第二次猜测。

    代码如下:

    #include <cstdio>
    
    #include <thrust/pair.h>
    
    #include "Utilities.cuh"
    
    #define BLOCKSIZE_MESHGRID_X    16
    #define BLOCKSIZE_MESHGRID_Y    16
    
    #define DEBUG
    
    /*******************/
    /* MESHGRID KERNEL */
    /*******************/
    template <class T>
    __global__ void meshgrid_kernel(const T * __restrict__ x, size_t Nx, const float * __restrict__ y, size_t Ny, T * __restrict__ X, T * __restrict__ Y) 
    {
        unsigned int tidx = blockIdx.x * blockDim.x + threadIdx.x;
        unsigned int tidy = blockIdx.y * blockDim.y + threadIdx.y;
    
        if ((tidx < Nx) && (tidy < Ny)) {   
            X[tidy * Nx + tidx] = x[tidx];
            Y[tidy * Nx + tidx] = y[tidy];
        }
    }
    
    /************/
    /* MESHGRID */
    /************/
    template <class T>
    thrust::pair<T *,T *> meshgrid(const T *x, const unsigned int Nx, const T *y, const unsigned int Ny) {
    
        T *X; gpuErrchk(cudaMalloc((void**)&X, Nx * Ny * sizeof(T)));
        T *Y; gpuErrchk(cudaMalloc((void**)&Y, Nx * Ny * sizeof(T)));
    
        dim3 BlockSize(BLOCKSIZE_MESHGRID_X, BLOCKSIZE_MESHGRID_Y);
        dim3 GridSize (iDivUp(Nx, BLOCKSIZE_MESHGRID_X), iDivUp(BLOCKSIZE_MESHGRID_Y, BLOCKSIZE_MESHGRID_Y));
    
        meshgrid_kernel<<<GridSize, BlockSize>>>(x, Nx, y, Ny, X, Y);
    #ifdef DEBUG
        gpuErrchk(cudaPeekAtLastError());
        gpuErrchk(cudaDeviceSynchronize());
    #endif
    
        return thrust::make_pair(X, Y);
    }
    
    /********/
    /* MAIN */
    /********/
    int main()
    {
        const int Nx = 3;
        const int Ny = 4;
    
        float *h_x = (float *)malloc(Nx * sizeof(float));
        float *h_y = (float *)malloc(Ny * sizeof(float));
    
        float *h_X = (float *)malloc(Nx * Ny * sizeof(float));
        float *h_Y = (float *)malloc(Nx * Ny * sizeof(float));
    
        for (int i = 0; i < Nx; i++) h_x[i] = i;
        for (int i = 0; i < Ny; i++) h_y[i] = i + 4.f;
    
        float *d_x; gpuErrchk(cudaMalloc(&d_x, Nx * sizeof(float)));
        float *d_y; gpuErrchk(cudaMalloc(&d_y, Ny * sizeof(float)));
    
        gpuErrchk(cudaMemcpy(d_x, h_x, Nx * sizeof(float), cudaMemcpyHostToDevice));
        gpuErrchk(cudaMemcpy(d_y, h_y, Ny * sizeof(float), cudaMemcpyHostToDevice));
    
        thrust::pair<float *, float *> meshgrid_pointers = meshgrid(d_x, Nx, d_y, Ny);
        float *d_X = (float *)meshgrid_pointers.first;
        float *d_Y = (float *)meshgrid_pointers.second;
    
        gpuErrchk(cudaMemcpy(h_X, d_X, Nx * Ny * sizeof(float), cudaMemcpyDeviceToHost));
        gpuErrchk(cudaMemcpy(h_Y, d_Y, Nx * Ny * sizeof(float), cudaMemcpyDeviceToHost));
    
        for (int j = 0; j < Ny; j++) {
            for (int i = 0; i < Nx; i++) {
                printf("i = %i; j = %i; x = %f; y = %f\n", i, j, h_X[j * Nx + i], h_Y[j * Nx + i]);
            }
        }
    
        return 0;
    
    }
    

    【讨论】:

      猜你喜欢
      • 1970-01-01
      • 2011-09-08
      • 2016-05-06
      • 2016-06-14
      • 2012-02-21
      • 2016-11-07
      • 1970-01-01
      • 2014-09-11
      • 2011-07-14
      相关资源
      最近更新 更多